from flask import Flask, request, current_app
import os
import logging
from logging.handlers import RotatingFileHandler

from utk_curio.backend.config import Config as config_class
from utk_curio.backend.extensions import db, migrate


def get_locale():
    return request.accept_languages.best_match(current_app.config['LANGUAGES'])


def create_app(config_class=config_class):

    # Flask app
    app = Flask(__name__)
    app.config.from_object(config_class)

    db.init_app(app)
    migrate.init_app(app, db)

    from utk_curio.backend.app.api import bp as api_bp
    app.register_blueprint(api_bp, url_prefix='')

    from utk_curio.backend.app.users import bp as users_bp
    app.register_blueprint(users_bp)

    if not app.debug and not app.testing:
        if app.config['LOG_TO_STDOUT']:
            stream_handler = logging.StreamHandler()
            stream_handler.setLevel(logging.INFO)
            app.logger.addHandler(stream_handler)
        else:
            if not os.path.exists('logs'):
                os.mkdir('logs')
            file_handler = RotatingFileHandler('logs/backend.log',
                                               maxBytes=10240, backupCount=10)
            file_handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)s: %(message)s '
                '[in %(pathname)s:%(lineno)d]'))
            file_handler.setLevel(logging.INFO)
            app.logger.addHandler(file_handler)

        app.logger.setLevel(logging.INFO)
        app.logger.info('Backend startup')

    return app



"""Rename profileImage to profile_image

Revision ID: 62a73b08f79f
Revises: 13c00188213d
Create Date: 2024-02-15 09:16:48.535205

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '62a73b08f79f'
down_revision = '13c00188213d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('profile_image', sa.String(length=200), nullable=True))
        batch_op.drop_column('profileImage')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('profileImage', sa.VARCHAR(length=200), nullable=True))
        batch_op.drop_column('profile_image')

    # ### end Alembic commands ###



import sqlite3
import sys
import os

def initialize_db(db_path="provenance.db"):
    # Ensure the directory exists
    os.makedirs(os.path.dirname(db_path), exist_ok=True) if os.path.dirname(db_path) else None

    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    commands = [
        "CREATE TABLE IF NOT EXISTS user (user_id integer primary key, user_name varchar(200), user_type varchar(20), user_IP varchar(20));",
        "CREATE TABLE IF NOT EXISTS versionTransaction (transaction_id integer primary key, transaction_date date, user_id integer, CONSTRAINT fk_trans_user FOREIGN KEY (user_id) REFERENCES user(user_id));",
        "CREATE TABLE IF NOT EXISTS version (version_id integer primary key, version_number varchar(10), transaction_id integer, CONSTRAINT fk_ver_trans FOREIGN KEY (transaction_id) REFERENCES versionTransaction(transaction_id_id));",
        "CREATE TABLE IF NOT EXISTS versionedElement (ve_id integer primary key, version_id integer, previous_ve_id integer, CONSTRAINT fk_ve_prevve FOREIGN KEY (previous_ve_id) REFERENCES versionedElement(ve_id), CONSTRAINT fk_ve_version FOREIGN KEY (version_id) REFERENCES version(version_id));",
        "CREATE TABLE IF NOT EXISTS workflow (workflow_id integer primary key, workflow_name varchar(20), ve_id integer, CONSTRAINT fk_wf_ve FOREIGN KEY (ve_id) REFERENCES versionedElement(ve_id));",
        "CREATE TABLE IF NOT EXISTS attribute (attribute_id integer primary key, attribute_name varchar(200), attribute_type varchar(20));",
        "CREATE TABLE IF NOT EXISTS relation (relation_id integer primary key, relation_name varchar(200));",
        "CREATE TABLE IF NOT EXISTS attributeRelation (ar_id integer PRIMARY KEY, attribute_id integer, relation_id integer, CONSTRAINT fk_ar_att FOREIGN KEY (attribute_id) REFERENCES attribute(attribute_id), CONSTRAINT fk_ar_rel FOREIGN KEY (relation_id) REFERENCES relation(relation_id));",
        "CREATE TABLE IF NOT EXISTS activity (activity_id integer primary key, workflow_id integer, ve_id integer, activity_name varchar(50), input_relation_id integer, output_relation_id integer, CONSTRAINT fk_act_wf FOREIGN KEY (workflow_id) REFERENCES workflow(workflow_id), CONSTRAINT fk_act_ve FOREIGN KEY (ve_id) REFERENCES versionedElement(ve_id), CONSTRAINT fk_act_rel_in FOREIGN KEY (input_relation_id) REFERENCES relation(relation_id), CONSTRAINT fk_act_rel_out FOREIGN KEY (output_relation_id) REFERENCES relation(relation_id));",
        "CREATE TABLE IF NOT EXISTS workflowExecution (workflowexec_id integer PRIMARY KEY, workflow_id integer, workflowexec_start_time time, workflowexec_end_time time, CONSTRAINT fk_wfe_wf FOREIGN KEY (workflow_id) REFERENCES workflow(workflow_id));",
        "CREATE TABLE IF NOT EXISTS relationInstance (re_id integer primary key, relation_id  integer, original_ri, CONSTRAINT fk_ri_rel FOREIGN KEY (relation_id) REFERENCES relation(relation_id), CONSTRAINT fk_ri_or FOREIGN KEY (original_ri) REFERENCES relationInstance(re_id));",
        "CREATE TABLE IF NOT EXISTS attributeValue (av_id integer primary key, attribute_id integer, ri_id integer, value varchar(200), CONSTRAINT fk_av_ri FOREIGN KEY  (ri_id) REFERENCES relationInstance(ri_id));",
        "CREATE TABLE IF NOT EXISTS activityExecution (activityexec_id integer PRIMARY KEY, activity_id integer, workflowexec_id integer, activityexec_start_time time, activityexec_end_time time, input_ri_id integer, output_ri_id integer, activity_source_code varchar(2000), CONSTRAINT fk_actexec_wfexec FOREIGN KEY (workflowexec_id) REFERENCES workflowExecution(workflowexec_id), CONSTRAINT fk_actexec_act FOREIGN KEY (activity_id) REFERENCES activity(activity_id), CONSTRAINT fk_actexec_rel_in FOREIGN KEY (input_ri_id) REFERENCES relationInstance(ri_id), CONSTRAINT fk_actexec_rel_out FOREIGN KEY (output_ri_id) REFERENCES relationInstance(ri_id));"
    ]

    for command in commands:
        cursor.execute(command)
        conn.commit()
    conn.close()

if __name__ == "__main__":
    db_path = sys.argv[1] if len(sys.argv) > 1 else "provenance.db"
    initialize_db(db_path)







import os
from dotenv import load_dotenv


basedir = os.path.abspath(os.path.dirname(__file__))
load_dotenv(os.path.join(basedir, '.env'))
load_dotenv(os.path.join(basedir, '.flaskenv'))

class Config:
    LOG_TO_STDOUT = os.environ.get('LOG_TO_STDOUT')

    # Flask-Caching related configs
    CACHE_TYPE = "FileSystemCache"
    CACHE_DEFAULT_TIMEOUT = 30000
    CACHE_IGNORE_ERRORS=True
    CACHE_DIR="./flask_cache"

    os.environ.setdefault('FLASK_APP', 'server.py')


from pyproj import Transformer
import xarray as xr
import json
from netCDF4 import Dataset
import numpy as np
import os
import math

# from wrf import getvar, interplevel, to_np, ALL_TIMES

# def thematic_from_netcdf(file_path, variables, coords, layer_id, operations=[], time_indexes=[], bbox={}):
#     def cdiff(scalar, axis=0):
#         '''
#         Performs the same as GrADS function cdiff()
#         http://cola.gmu.edu/grads/gadoc/gradfunccdiff.html
#         The scalar quantity must by 2D.
#         The finite differences calculation ignores the borders, where np.nan is returned.
#         '''
#         # Check if 2D
#         dimScalar = np.size(np.shape(scalar))
#         if dimScalar != 2:
#             print(
#                 "Pystuff Error: scalar must have only 2 dimensions, but it has %d." % dimScalar)
#             return

#         # Length of each dimension
#         lendim0 = np.shape(scalar)[0]
#         lendim1 = np.shape(scalar)[1]

#         # Initialize output var
#         out = np.zeros(np.shape(scalar))
#         out.fill(np.nan)

#         # Centered finite differences
#         for x in np.arange(1, lendim0-1):
#             for y in np.arange(1, lendim1-1):
#                 if axis == 0:
#                     out[x, y] = scalar[x+1, y]-scalar[x-1, y]
#                 elif axis == 1:
#                     out[x, y] = scalar[x, y+1]-scalar[x, y-1]
#                 else:
#                     print(
#                         "Pystuff Error: Invalid axis option. Must be either 0 or 1.")
#                     return
#         return out
    
#     def hdivg(_u, _v, _lat, _lon, _r=6.371*(10**6)):
#         '''
#         Calculates the horizontal divergence (du/dx+dv/dy) exactly like GrADS. 
#         lat and lon are 1D arrays.
#         http://cola.gmu.edu/grads/gadoc/gradfunccdiff.html
#         '''
#         latv, lonv = np.meshgrid(_lat, _lon, indexing='ij')

#         # r = 
#         dtr = np.pi/180
#         dudx = cdiff(_u, axis=1)/cdiff(lonv*dtr, axis=1)
#         dvdy = cdiff(_v*np.cos(latv*dtr), axis=0)/cdiff(latv*dtr, axis=0)
#         out = (dudx + dvdy)/(_r*np.cos(latv*dtr))
#         return out
    
#     ncData = Dataset(file_path)
    
#     lat_arr = []
#     lon_arr = []

#     if len(ncData.variables[coords['lat']].shape) == 1:
#         lat_arr = ncData.variables[coords['lat']]
    
#     else:
#         lat_matrix = ncData.variables[coords['lat']][:][0]
#         lat_arr = np.array([row[0] for row in lat_matrix] )

#     if len(ncData.variables[coords['lon']].shape) == 1:
#         lon_arr = ncData.variables[coords['lon']]
    
#     else:
#         lon_arr = ncData.variables[coords['lon']][:][0][0]

#     min_lat_idx = 0
#     max_lat_idx = len(lat_arr)-1
        
#     min_lon_idx = 0
#     max_lon_idx = len(lon_arr)-1

#     if bbox:
#         min_lat_idx = None
#         max_lat_idx = None
            
#         min_lon_idx = None
#         max_lon_idx = None

#         for i in range(len(lat_arr)):
#             if min_lat_idx is None and lat_arr[i] >= bbox['min_lat'] >=bbox['min_lat']:
#                 min_lat_idx = i

#             elif max_lat_idx is None and lat_arr[i] >= bbox['max_lat'] >=bbox['max_lat']:
#                 max_lat_idx = i
#                 break

#         for i in range(len(lon_arr)):
#             if min_lon_idx is None and lon_arr[i] >= bbox['min_lon'] >=bbox['min_lon']:
#                 min_lon_idx = i

#             elif max_lon_idx is None and lon_arr[i] >= bbox['max_lon'] >=bbox['max_lon']:
#                 max_lon_idx = i
#                 break


#         # lat_idxs = [i for i in range(len(lat_arr)) if lat_arr[i] >= bbox['min_lat'] and lat_arr[i] <= bbox['max_lat']]
#         # lon_idxs = [j for j in range(len(lon_arr)) if lon_arr[j] >= bbox['min_lon'] and lon_arr[j] <= bbox['max_lon']]
    
#         # min_lat_idx = lat_idxs[0]
#         # max_lat_idx = lat_idxs[-1]
        
#         # min_lon_idx = lon_idxs[0]
#         # max_lon_idx = lon_idxs[-1]

# ############################################################################################
    
#     # data = ncData.variables[variables[0]['name']][:]
#     # new_data = np.array(data, copy=True)

#     # # addition = lambda a, b: a + b
#     # # multiplication = lambda a, b: a * b
#     # # division = lambda a, b: a / b

#     #############################################################################
#     final_data = []

#     for operation in operations:
#         if operation['type'] == 'vector':
#             if operation['dimension'] == 'space':
#                 if operation['function'] == 'sum':
#                     for v in variables:
#                         var_name = v['name']
#                         final_data = ncData.variables[v['name']][:] if len(final_data) == 0 else final_data + ncData.variables[v['name']][:]

#                 elif operation['function'] == 'interpolation':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']
                    
#                     var_name = [v['name'] for v in variables if v['key'] != 'pressure'][0]

#                     time_idxs = list(range(5))
#                     level = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0'][0]

#                     for tidx in time_idxs:
#                         pressure_matrix = getvar(ncData, pressure_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         var_matrix = getvar(ncData, var_name, timeidx=tidx,
#                         squeeze=False, meta=False)
                        
#                         var_matrix = to_np(interplevel(var_matrix, pressure_matrix, level))

#                         if np.ma.isMaskedArray(var_matrix):
#                             var_matrix = list(var_matrix)
#                         final_data.append(var_matrix)

#                     final_data = np.array(final_data)

#                 elif operation['function'] == 'calculate_wind':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']

#                     u_name = [v['name'] for v in variables if v['key'] == 'u'][0]
#                     v_name = [v['name'] for v in variables if v['key'] == 'v'][0]

#                     time_idxs = list(range(5))
#                     level = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0'][0]

#                     for tidx in time_idxs:
#                         pressure = getvar(ncData, pressure_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         u = getvar(ncData, u_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         v = getvar(ncData, v_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         # to do

#                 elif operation['function'] == 'calculate_hdiv':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']

#                     u_name = [v['name'] for v in variables if v['key'] == 'u'][0]
#                     v_name = [v['name'] for v in variables if v['key'] == 'v'][0]

#                     time_idxs = list(range(5))
#                     level = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0'][0]

#                     for t in range(len(time_idxs)):
#                         u = getvar(ncData, u_name, timeidx=t, squeeze=False, meta=False)
#                         v = getvar(ncData, v_name, timeidx=t, squeeze=False, meta=False)
#                         pressure = getvar(ncData, pressure_name, timeidx=t,
#                                 squeeze=False, meta=False)
                        
#                         u_interp = to_np(interplevel(u, pressure, level))
#                         v_interp = to_np(interplevel(v, pressure, level))

#                         hdiv = hdivg(u_interp, v_interp, lat_arr, lon_arr)
#                         final_data.append(hdiv)
                             
#                 elif operation['function'] == 'calculate_kidx':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']

#                     tc_name = [v for v in variables if v['key'] == 'tc'][0]
#                     rh_name = [v for v in variables if v['key'] == 'rh'][0]

#                     level_0 = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0' ]
#                     level_1 = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_1' ]
#                     level_2 = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_2' ]

#                     a = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'a']
#                     b = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'b']
#                     mv = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'mv']
#                     le = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'le']
#                     r = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'r']

#                     for t in range(len(time_idxs)):
                    
#                         tc = getvar(ncData, tc_name, timeidx=t, squeeze=False, meta=False)
#                         rh = getvar(ncData, rh_name, timeidx=t, squeeze=False, meta=False)
#                         pressure = getvar(ncData, pressure_name, timeidx=t, squeeze=False, meta=False)

#                         tc_interp_0 = to_np(interplevel(tc, pressure, level_0))
#                         tc_interp_1 = to_np(interplevel(tc, pressure, level_1))
#                         tc_interp_2 = to_np(interplevel(tc, pressure, level_2))

#                         rh_interp_0 =  to_np(interplevel(tc, pressure, level_0))
#                         rh_interp_1 =  to_np(interplevel(tc, pressure, level_1))

#                         arr = []

#                         for i in range(len(lat_arr)):
#                             row_arr = []

#                             for j in range(len(lon_arr)):
#                                 t1 = tc_interp_0[i][j]
#                                 t2 = tc_interp_1[i][j]
#                                 t4 = tc_interp_2[i][j]

#                                 rh_0 = rh_interp_0[i][j]
#                                 rh_1 = rh_interp_1[i][j]

#                                 e3 = a * math.exp((mv*le/r) * ((1/b) - (1/(t1+b))))
#                                 e5 = a * math.exp((mv*le/r) * ((1/b) - (1/(t4+b))))
                                
#                                 es3 = rh_0[i][j]*e3/100
#                                 es5 = rh_1[i][j]*e5/100

#                                 td3 = 1/(1/b-(r/(mv*le)) * math.log10(es3/a)) - k
#                                 td5 = 1/(1/b-(r/(mv*le)) * math.log10(es5/a)) - k

#                                 t3 = td3
#                                 t5 = td5

#                                 k = t1-t2+t3-t4+t5

#                                 row_arr.append(k)
#                             arr.append(row_arr)
#                         final_data.append(arr)
#                     final_data = np.array(final_data)
            
#             elif operation['dimension'] == 'time':
#                 if operation['function'] == 'accumulate':
#                     interval = operation['parameters'][0]
#                     final_data = [final_data[t] + final_data[t + interval] for t in range(0, len(final_data), interval)]
                
#                 elif operation['function'] == 'unaccumulate':
#                     interval = operation['parameters'][0]
#                     final_data = [final_data[t] - final_data[t - interval] for t in range(len(final_data), 0-1, interval)]

#                 elif operation['function'] == 'avg':
#                     final_data = [final_data[t] - final_data[t - interval] for t in range(len(final_data), 0-1, interval)]

#         elif operation['type'] == 'vector_to_scalar':
#             if len(final_data) == 0:
#                 final_data = ncData.variables[variables[0]['name']][:]
            
#             if operation['function'] == 'avg':
#                 func = lambda x: np.mean(x)
#                 arr = [func(matrix) for matrix in final_data]

#                 final_data = np.array(arr)
        
#         elif operation['type'] == 'scalar':
#             if len(final_data) == 0:
#                 new_data = ncData.variables[variables[0]['name']][:]
            
#             for matrix in new_data:
#                 for row in matrix:
#                     for i in range(len(row)):
#                         apply_scalar = eval(operation['function'])
#                         row[i] = float(round(apply_scalar(row[i], operation['value']), 2))


#     # rainc = ncData.variables['RAINC'][:]
#     # rainnc = ncData.variables['RAINNC'][:]
    
#     # print(rainc[4][0][0], rainnc[4][0][0], new_data[4][0][0])

    

############################################################################################


    # points = []
    # values = []

    # for latidx in range(min_lat_idx, max_lat_idx + 1):
    #     for lonidx in range(min_lon_idx, max_lon_idx + 1):
    #         points.append((lat_arr[latidx], lon_arr[lonidx]))
    #         pt_arr = []
            
    #         for tidx in range(len(data)):
    #             pt_arr.append(float(data[tidx][latidx][lonidx]))

    #         values.append(pt_arr)

    # coordinates = []
    # transformer = Transformer.from_crs(coords['proj'], 3395)

    # for point in transformer.itransform(points):
    #     coordinates.append(float(point[0]))
    #     coordinates.append(float(point[1]))
    #     coordinates.append(0)

    # abstract_json = {
    #     "id": layer_id,
    #     "coordinates": coordinates,
    #     "values": values
    # }

    # json_object = json.dumps(abstract_json)

    # directory = os.path.dirname(file_path)

    # with open(os.path.join(directory,layer_id+".json"), "w") as outfile:
    #     outfile.write(json_object)
    

  
    
    
    # if len(variables'] == 1):
    #     var_name = variables'][0]['name']
    #     var_unit = variables'][0]['unit']
    #     var_level = variable'][0]['level']
    #     var_key   = variable'][0]['key']


if __name__ == '__main__':   
    # myobj = json.load(open("./netcdf_grammarv1.json") )


    file_path = "../../examples/wrf_chicago/wrfout_d03_2023-06-09_20_00_00"
    variables = [ #'RAINC', 'RAINNC'
        # {
        #     "name": "T2",
        #     "level": "",
        #     "unit": "C",
        #     "key": ""
        # }
            {
                "name": 'pressure',
                "key": 'pressure',
            },
            {
                "name": 'u',
                "key": 'u',
            },
            {
                "name": 'v',
                "key": 'v',
            }

    ]

    coords = {
        "lat": "XLAT",
        "lon": "XLONG",
        "proj": 4326,
    }

    bbox = {
            "min_lat": 33,
            "min_lon": -94,
            "max_lat": 42,
            "max_lon": -81
        }

    # operations = {
    #         "scalar": [ 
    #             # {
    #             #     "name": "addition",
    #             #     "value": -273,
    #             # },
    #             # {
    #             #     "name": "multiplication",
    #             #     "value": 2,
    #             # },
    #             # {
    #             #     "name": "division",
    #             #     "value": 3,
    #             # },
    #             {
    #                 "name": "lambda x, l: (x + l[0]) * l[1]",
    #                 "value": [ -273, 2],
    #             }

    #         ],

    #         "vector": [
    #             # {
    #             #     "name": "",
    #             #     "parameters" : {
    #             #         "name": "",
    #             #         "value": ""

    #             #     }
    #             # }
    #         ]
    #     }

    # operations = {
    #     "scalar": {
    #         "function": "lambda x, l: (x + l[0]) * l[1]",
    #         "parameters": [-273, 2]
    #     },

    #     "vector" : {
    #         "type": "between_variables",
    #         "function": "lambda x, y: x + y",
    #         "parameters": [],
    #     }
        
        
    # }

    operations = [
        {
            "type": "vector",
            "dimension": "space",
            "function": "calculate_wind",
            "parameters": [
                {
                    "name": "level",
                    "value": 850
                }

            ],
        },
        # {
        #     "type": "vector",
        #     "dimension": "space",
        #     "function": "sum",
        #     "parameters": [],
        # },
        # {
        #     "type": "scalar",
        #     "function": "lambda x, l: (x + l[0]) * l[1]",
        #     "parameters": [-273, 2]
        # },
    ]

    time_indexes = []
    layer_id = "wrfout_d03_2023-06-09_20_00_00_T2"

    # thematic_from_netcdf(file_path, variables, coords, operations, time_indexes, bbox)
    # thematic_from_netcdf(file_path, variables, coords, layer_id, operations)

    # def test():
        

    # test = lambda : print('hey')
    # a = eval('test')

    # a()
    
    # addMatrix = np.arange(3.0)
    # print(addMatrix)

# add_matrix = np.full((2, 3, 5), 7)
# print(add_matrix)


"""Create User

Revision ID: 8273cbb189e4
Revises: 
Create Date: 2024-02-14 18:41:16.158811

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8273cbb189e4'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('user',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('email', sa.String(length=120), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('profileImage', sa.String(length=200), nullable=True),
    sa.Column('type', sa.String(length=100), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('email')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('user')
    # ### end Alembic commands ###



import os
from setuptools import setup, find_packages
from distutils.dir_util import copy_tree

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

setup(
    name="utk",
    version="0.8.9",
    author="Fabio Miranda",
    author_email="fabiom@uic.edu",
    description="The Urban Toolkit: A Grammar-based Framework for Urban Visual Analytics",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/urban-toolkit/utk/',
    packages=['utk'],
    package_dir = {
        'utk': './src/utk'
    },
    package_data={'utk': ['../utk_server.py', '../utk-frontend/build/utk-app/*', '../../requirements.txt']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'utk = utk_server:main',
        ],
    },
    license_files = ('LICENSE'),
    python_requires='>=3.9',
    install_requires=install_requires
)



import pysolar
import threading
import pytz
import struct

import timezonefinder
import math
import numpy as np

from datetime import datetime
from datetime import timedelta

import json
import os
import sys

if sys.platform != "darwin":
    from plotoptix import NpOptiX
    from plotoptix.geometry import PinnedBuffer

class ShadowAccumulator:
    '''
        Calculate shadow accumulation considering meshes stored in json files.
    '''

    data = []
    # start = None
    # end = None
    intervals = [] # list of lists containing different time intervals

    flat_coords = []
    coords = np.array([])
    indices = np.array([])
    ids = np.array([])
    normals = np.array([])
    ids_per_structure = [] # the ids are local per structure. They have to be globalized later
    coords_per_file = [] # stores the number of coordinates per file to write the shadow data back to the correct files
    coords_before_transformation = []
    per_face_avg_accum = []
    latitude = 0
    longitude = 0
    result_to_write = {} # the result that will be outputed in the files (groupped by interval index)

    def __init__(self, latitude, longitude, data, intervals):

        '''
            All meshes must be 3D

            * @param {List[string]} filespaths All the layers containing meshes that have to be considered in the shadow calculation
            * @param {string} start Timestamp of the beginning of the accumulation. Format: "%m/%d/%Y %H:%M". Example: "03/20/2015 10:00"
            * @param {string} end Timestamp of the end of the accumulation. Format: "%m/%d/%Y %H:%M". Example: "03/20/2015 11:01"
        '''

        for interval in intervals:
            start = datetime.strptime(interval[0], "%m/%d/%Y %H:%M")
            end = datetime.strptime(interval[1], "%m/%d/%Y %H:%M")

            self.intervals.append([start, end])

        self.data = data
        # self.start = datetime.strptime(start, "%m/%d/%Y %H:%M")
        # self.end = datetime.strptime(end, "%m/%d/%Y %H:%M")

        self.latitude = latitude
        self.longitude = longitude

    def computeVector(self, alt, azm):
        alt = math.pi*alt/180.0
        azm = math.pi/2.0-math.pi*azm/180.0

        x = math.cos(alt)*math.cos(azm)
        y = math.cos(alt)*math.sin(azm)
        z = math.sin(alt)

        nrm = math.sqrt(x*x+y*y+z*z)
    #     nrm = nrm if nrm>sys.float_info.epsilon else 1

        return [x/nrm,y/nrm,z/nrm]

    def computeAngle(self, vec1, vec2):
        # temporary fix for [0,0,0] normals TODO
        if vec1[0] == 0 and vec1[1] == 0 and vec1[2] == 0:
            vec1 = np.array([0.0,0.0,1.0])

        unit_vector_1 = vec1 / np.linalg.norm(vec1)
        unit_vector_2 = vec2 / np.linalg.norm(vec2)
        dot_product = np.dot(unit_vector_1, unit_vector_2)
        angle = np.arccos(dot_product) * 180.0 / math.pi #degrees)

        return angle

    # computer all directions of sun every nskip between start and end
    def compute_directions(self, start, end, lat, lng, nskip=1):

        tf = timezonefinder.TimezoneFinder()
        tz = tf.timezone_at(lng=lng, lat=lat)
        tz = pytz.timezone(tz)
        start = tz.localize(start)
        end = tz.localize(end)

        directions = []

        curr = start
        until = end
        delta=timedelta(minutes=nskip)
        directions = []
        while curr < until:
            curr += delta
            alt = pysolar.solar.get_altitude(lat, lng, curr) # angle between the sun and a plane tangent to the earth at lat/lng
            azm = (pysolar.solar.get_azimuth(lat, lng, curr))
            ddir = self.computeVector(alt, azm)
            directions.append(ddir)
        return directions

    class params:
        done = threading.Event()
        k = 0

    # computes the shadow accumulation 
    def compute(self, directions, coords, indices, normals):
        def done(rt: NpOptiX) -> None:
            self.params.k += 1
            self.params.done.set()

        camera_plane_dim = math.ceil(math.sqrt(coords.shape[0]))
        width = camera_plane_dim # camera plane width
        height = camera_plane_dim # camera plane height

        accumulation = np.full((coords.shape[0], 1), 0) 

        rt = NpOptiX(width=width, height=height)
        rt.set_mesh('buildings', pos=coords, faces=indices, normals=normals)#, c=colors)
        rt.set_float("scene_epsilon", 0.01) # set shader variable with a given name
        rt.set_param(min_accumulation_step=1, max_accumulation_frames=1) # set raytracer parameter(s)
        rt.set_accum_done_cb(done)
        rt.start()

        for direction in directions:
            
            with PinnedBuffer(rt.geometry_data['buildings'], 'Positions') as P:
                n = math.ceil(math.sqrt(len(P)))
                eye = np.zeros((n,n,4), dtype=np.float32) 

                eyeRow = -1
                for index, elem in enumerate(P):
                    if(index%n == 0):
                        eyeRow += 1
                    eye[eyeRow,index%n,:3] = elem + 1e-1 * normals[index]
    
                rt.set_texture_2d('eye', eye, refresh=True)

            with PinnedBuffer(rt.geometry_data['buildings'], 'Vectors') as N: 
                n = len(normals)

                rdir_temp = []
                for i in range(0, n):
                    normal = normals[i]
                    ang = self.computeAngle(normal, direction)
                    if ang > 90.0:
                        rdir_temp.append([0,0,0,0])
                    else:
                        rdir_temp.append([direction[0],direction[1],direction[2],-1])
                rdir_temp = np.array(rdir_temp, dtype=np.float32)

                n = math.ceil(math.sqrt(len(rdir_temp)))
                rdir = np.zeros((n,n,4), dtype=np.float32) 

                rdirRow = -1
                for index, elem in enumerate(rdir_temp):
                    if(index%n == 0):
                        rdirRow += 1
                    rdir[rdirRow,index%n, :] = elem


                rt.set_texture_2d('dir', rdir, refresh=True)

            self.params.done.clear() # resetting the thread flag       
            rt.setup_camera('cam2', cam_type='CustomProjXYZtoDir', textures=['eye', 'dir'], make_current=True) # two 4D textures are defined ([height, width, 4]). 'eye' is composed of origin points ([x, y, z, 0]). 'dir' is composed of ray directions and maximum ranges ([cosx, cosy, cosz, r]).

            self.params.done.wait() # wait for the ray tracer to finish
            
            dist = rt._hit_pos[:,:,3].reshape(rt._height, rt._width) # _hit_pos shape: (height, width, 4). This 4 refers to [X, Y, Z, D], where XYZ is the hit 3D position and D is the hit distance to the camera plane. We are only interested in the D.
            dist = [item for sublist in dist for item in sublist] # flattening distance 
            dist = dist[:coords.shape[0]] # dropping extra points in the end of the matrix 
            dist = np.array(dist) 

            dist[dist < 0xFFFFFFFF] = 1
            dist[dist > 0xFFFFFFFF] = 0

            accumulation[:,0] = accumulation[:,0]+dist

        rt.close()

        return accumulation

    def per_face_avg(self, accumulation, indices, ids, ids_per_buildings):

        #make the ids global
        global_ids = []

        current_building = 0
        read_ids = 0
        offset = 0
        for i in range(0,len(ids)):
            if read_ids == ids_per_buildings[current_building]:
                read_ids = 0
                current_building += 1
                offset = len(global_ids)
            global_ids.append(ids[i]+offset)
            read_ids += 1

        avg_accumulation_triangle = np.zeros(len(indices))
        avg_accumulation_cell = np.zeros(len(global_ids))

        # calculate acc by triangle
        for i in range(0,len(indices)):
            value = 0
            for vid in indices[i]:
                value += accumulation[int(vid)]
            avg_accumulation_triangle[i] = value

        # calculate acc by cell based on the triangles that compose it
        count_acc_cell = np.zeros(len(global_ids))
        for i in range(0, len(indices)):
            cell = int(global_ids[i])       
            avg_accumulation_cell[cell] += avg_accumulation_triangle[i]
            count_acc_cell[cell] += 1

        # distribute the average of the cell to the triangles that compose it
        for i in range(0, len(indices)):
            cell = int(global_ids[i])       
            avg_accumulation_triangle[i] = avg_accumulation_cell[cell]/count_acc_cell[cell]    

        return np.array(avg_accumulation_triangle)

    def accumulate(self, start, end, lat, lng, coords, indices, normals, nskip=1):
        directions = self.compute_directions(start, end, lat, lng, nskip)

        accumulation = self.compute(directions, coords, indices, normals)
        return accumulation

    def per_coordinates_avg(self, avg_accumulation_triangle, coords, indices):
        '''
            Distributes triangle avg to the coordinates that composes the triangle. The coordinates need to be duplicated, meaning that there are unique indices. 
        '''

        avg_accumulation_per_coordinates = np.zeros(len(coords), dtype=np.float32) 

        for index, elem in enumerate(avg_accumulation_triangle):
            avg_accumulation_per_coordinates[indices[index][0]] = elem
            avg_accumulation_per_coordinates[indices[index][1]] = elem
            avg_accumulation_per_coordinates[indices[index][2]] = elem

        return np.array(avg_accumulation_per_coordinates)

    def save(self):
        '''
            Writes the shadow data back to the mesh files passed to the constructor
        '''

        for index in range(len(self.intervals)):

            accumulation = self.result_to_write[index] # shadow data accumulated per coordinates
            function_index = index

            # function values are the normalized accumulation values ([0,1]) that are used by the shader of utk-map to color the cells
            if(max(accumulation) != 0):
                function_values = accumulation/max(accumulation)
            else:
                function_values = accumulation

            function_values = function_values.tolist()

            flat_coords = self.flat_coords.copy()

            # decoupled abstract layer
            # shadow_layer = {'id': "shadow"+str(function_index), 'coordinates': [round(item,4) for item in self.flat_coords], 'values': [round(item,4) for item in function_values]}

            # directory = os.path.dirname(self.filespaths[0])

            # with open(os.path.join(directory, "shadow"+str(function_index)+".json"), "w") as outfile:
            #     json.dump(shadow_layer, outfile)

            for index, geometries_count in enumerate(self.coords_per_file):
                
                fileName = ""

                if(type(self.data[index]) == str): # file path
                    fileName = os.path.splitext(os.path.basename(self.data[index]))[0]
                else: # the raw data
                    fileName = self.data[index]["id"]

                directory = "./"

                if(type(self.data[index]) == str): # file path
                    directory = os.path.dirname(self.data[index])

                function_values_this_file = []
                flat_coords_this_file = []

                for geometry_count in geometries_count:

                    function_values_this_file += function_values[:geometry_count]

                    flat_coords_this_file += flat_coords[:geometry_count*3]

                    flat_coords = flat_coords[geometry_count*3:] # remove the values that belong to the current mesh
                    function_values = function_values[geometry_count:] # remove the values that belong to the current mesh

                shadow_layer = {'id': "shadow"+str(function_index)+'_'+fileName, 'coordinates': [round(item,4) for item in flat_coords_this_file], 'values': [round(item,4) for item in function_values_this_file]}

                with open(os.path.join(directory, "shadow"+str(function_index)+'_'+fileName+".json"), "w") as outfile:
                    json.dump(shadow_layer, outfile)

    # returns a dictionary with shadow values separated by layers and groupped by object
    def get_shadow_by_layer(self):
        
        returned_dict = {}

        for index in range(len(self.intervals)):

            accumulation = self.result_to_write[index] # shadow data accumulated per coordinates
            function_index = index

            # function values are the normalized accumulation values ([0,1]) that are used by the shader of utk-map to color the cells
            if(max(accumulation) != 0):
                function_values = accumulation/max(accumulation)
            else:
                function_values = accumulation

            function_values = function_values.tolist()

            flat_coords = self.flat_coords.copy()

            for index, geometries_count in enumerate(self.coords_per_file):
                
                fileName = ""

                if(type(self.data[index]) == str): # file path
                    raise Exception("get_shadow_by_layer can only be used when raw data and not filepaths are provided")
                else: # the raw data
                    fileName = self.data[index]["id"]

                function_values_this_file = []
                coords_this_file = []

                for geometry_count in geometries_count:

                    function_values_this_file.append([round(item, 4) for item in function_values[:geometry_count]])

                    coords_this_file.append(flat_coords[:geometry_count*3])

                    flat_coords = flat_coords[geometry_count*3:] # remove the values that belong to the current mesh
                    function_values = function_values[geometry_count:] # remove the values that belong to the current mesh

                shadow_layer = {'id': "shadow"+str(function_index)+'_'+fileName, 'coordinates': coords_this_file, 'values': function_values_this_file}

                returned_dict["shadow"+str(function_index)+'_'+fileName] = shadow_layer
        
        return returned_dict

    def accumulate_shadow(self):
        '''
            Accumulate shadow over a period of time considering the parameters defined in the constructor
        '''

        self.load_files()

        self.flat_coords = [float(elem) for sublist in self.coords_before_transformation for elem in sublist]

        for index, interval in enumerate(self.intervals):
            accum = self.accumulate(interval[0], interval[1], self.latitude, self.longitude, self.coords, self.indices, self.normals, 15)

            self.per_face_avg_accum = self.per_face_avg(accum, self.indices, self.ids, self.ids_per_structure) # accumulation per triangle

            max = -1
            min = -1

            for elem in accum:

                if max == -1:
                    max = elem[0]

                if min == -1:
                    min = elem[0]

                if elem[0] > max:
                    max = elem[0]

                if elem[0] < min:
                    min = elem[0]
 
            self.result_to_write[index] =  [(elem[0] - min) / (max - min) for elem in accum]

    def load_files(self):

        for value in self.data:

            file_content = {}

            if(type(value) == str): # filepath
                filepath = value

                file = open(filepath, mode='r')
                file_content = json.loads(file.read())

                directory = os.path.dirname(filepath)
                # file name with extension
                file_name = os.path.basename(filepath)
                # file name without extension
                file_name_wo_extension = os.path.splitext(file_name)[0]

                coordinates = []
                normals = []
                indices = []
                ids = []

                if('coordinates' in file_content['data'][0]['geometry']):
                    f = open(os.path.join(directory,file_name_wo_extension+'_coordinates.data'), "rb")

                    data = f.read()

                    unpacked_data = struct.iter_unpack('d', data)

                    for elem in unpacked_data:
                        coordinates.append(elem[0])

                    f.close()
                if('normals' in file_content['data'][0]['geometry']):
                    f = open(os.path.join(directory,file_name_wo_extension+'_normals.data'), "rb")

                    data = f.read()

                    unpacked_data = struct.iter_unpack('f', data)

                    for elem in unpacked_data:
                        normals.append(elem[0])

                    f.close()
                if('indices' in file_content['data'][0]['geometry']):
                    f = open(os.path.join(directory,file_name_wo_extension+'_indices.data'), "rb")

                    data = f.read()

                    unpacked_data = struct.iter_unpack('I', data)

                    for elem in unpacked_data:
                        indices.append(elem[0])

                    f.close()
                if('ids' in file_content['data'][0]['geometry']):
                    f = open(os.path.join(directory,file_name_wo_extension+'_ids.data'), "rb")

                    data = f.read()

                    unpacked_data = struct.iter_unpack('I', data)

                    for elem in unpacked_data:
                        ids.append(elem[0])

                    f.close()

                for i in range(len(file_content['data'])):

                    if(len(coordinates) > 0):
                        startAndSize = file_content['data'][i]['geometry']['coordinates']
                        file_content['data'][i]['geometry']['coordinates'] = coordinates[startAndSize[0]:startAndSize[0]+startAndSize[1]]

                    if(len(indices) > 0):
                        startAndSize = file_content['data'][i]['geometry']['indices']
                        file_content['data'][i]['geometry']['indices'] = indices[startAndSize[0]:startAndSize[0]+startAndSize[1]]

                    if(len(normals) > 0):
                        startAndSize = file_content['data'][i]['geometry']['normals']
                        file_content['data'][i]['geometry']['normals'] = normals[startAndSize[0]:startAndSize[0]+startAndSize[1]]

                    if(len(ids) > 0):
                        startAndSize = file_content['data'][i]['geometry']['ids']
                        file_content['data'][i]['geometry']['ids'] = ids[startAndSize[0]:startAndSize[0]+startAndSize[1]]
                
                file.close()
            else:
                file_content = value

            file_coords = []
            file_indices = []
            file_ids = []
            file_normals = []

            self.coords_per_file.append([])

            for element in file_content['data']:
                file_indices += [int(item)+int(len(file_coords)/3)+self.coords.shape[0] for item in element['geometry']['indices']] # concat in the end of the list # considers always a 3d mesh
                file_coords += element['geometry']['coordinates'] # concat in the end of the list
                file_ids += element['geometry']['ids'] # concat in the end of the list
                file_normals += element['geometry']['normals'] # concat in the end of the list

                self.ids_per_structure.append(len(element['geometry']['ids']))

                self.coords_per_file[-1].append(int(len(element['geometry']['coordinates'])/3)) # considers always a 3d mesh

            file_coords = np.reshape(np.array(file_coords), (int(len(file_coords)/3), -1)) # considers always a 3d mesh
            file_indices = np.reshape(np.array(file_indices), (int(len(file_indices)/3), -1)) # considers always a 3d mesh
            file_ids = np.array(file_ids)
            file_normals = np.reshape(np.array(file_normals), (int(len(file_normals)/3), -1)) # considers always a 3d mesh

            if len(self.coords) == 0:
                self.coords = np.copy(file_coords)
            else:
                self.coords = np.concatenate((self.coords, file_coords), axis=0)

            if len(self.indices) == 0:
                self.indices = np.copy(file_indices)
            else:
                self.indices = np.concatenate((self.indices, file_indices), axis=0)

            if len(self.ids) == 0:
                self.ids = np.copy(file_ids)
            else:
                self.ids = np.concatenate((self.ids, file_ids), axis=0)

            if len(self.normals) == 0:
                self.normals = np.copy(file_normals)
            else:
                self.normals = np.concatenate((self.normals, file_normals), axis=0)

        self.ids_per_structure = np.array(self.ids_per_structure)

        self.coords_before_transformation = np.copy(self.coords)

        self.coords = self.coords - np.mean(self.coords, axis=0)

    # def view(self):
    #     # if(max(self.per_face_avg_accum) != 0):
    #     #     ccolors = 255*plt.cm.YlOrRd((self.per_face_avg_accum/max(self.per_face_avg_accum)))
    #     # else:
    #     #     ccolors = 255*plt.cm.YlOrRd(self.per_face_avg_accum)

    #     vmesh = vedo.Mesh([self.coords, self.indices])

    #     # pts_hit_pos = vedo.Points(not_in_inf, c=(0, 128, 255))
    #     # arrows_normals = vedo.Arrows(self.coords, (4*self.normals+self.coords), thickness=2, c="green")

    #     # arrows_rdir = vedo.Arrows(eye + 1e-1 * normals, (rdir+(eye + 1e-1 * normals)), thickness=2, c="blue")

    #     # vmesh.cellIndividualColors(ccolors)
    #     vmesh.lineWidth(1.5)

    #     vplt = vedo.Plotter()
    #     vplt += vmesh.clone()
    #     # vplt += pts_hit_pos.clone()
    #     # vplt += arrows_normals.clone()
    #     # vplt += arrows_rdir.clone()
    #     vplt.show(viewup='z', zoom=1.3)


from flask import request, jsonify, g
from functools import wraps
from utk_curio.backend.app.users.models import UserSession, User


def require_auth(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        token = request.headers.get('Authorization')
        if not token:
            return jsonify({'error': 'Authorization token is missing'}), 401

        session = UserSession.query.filter_by(token=token, active=True).first()
        if not session:
            return jsonify({'error': 'Invalid or expired token'}), 403

        user = User.query.filter_by(id=session.user_id).first()
        if not user:
            return jsonify({'error': 'User not found'}), 404

        g.user = user  # Attach user to global g object
        return f(*args, **kwargs)
    return decorated_function


"""Add provider and provider_uid to User

Revision ID: 13c00188213d
Revises: f5c32a160bb4
Create Date: 2024-02-15 09:13:27.568691

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '13c00188213d'
down_revision = 'f5c32a160bb4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('provider', sa.String(length=50), nullable=True))
        batch_op.add_column(sa.Column('provider_uid', sa.String(length=200), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.drop_column('provider_uid')
        batch_op.drop_column('provider')

    # ### end Alembic commands ###



import unittest
import tempfile
import os
import sys
import json
from utk_curio.sandbox.app import app
from flask import Flask, jsonify

# Modify sys.path to include the sandbox folder
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'sandbox')))


class TestSandbox(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.client = app.test_client()

    def test_live(self):
        response = self.client.get('/live')
        self.assertEqual(response.data.decode('utf-8'), 'Sandbox is live.')
        self.assertEqual(response.status_code, 200)


if __name__ == "__main__":
    unittest.main()


from utk_curio.backend.extensions import db
from uuid import uuid4

class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(120), unique=True, nullable=False)
    name = db.Column(db.String(100), nullable=False)
    profile_image = db.Column(db.String(200), nullable=True)
    type = db.Column(db.String(100), nullable=True)
    provider = db.Column(db.String(50), nullable=True)
    provider_uid = db.Column(db.String(200), nullable=True)

    def __repr__(self):
        return '<User %r>' % self.email

class UserSession(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)
    token = db.Column(db.String(36), unique=True, default=lambda: str(uuid4()), nullable=False)
    active = db.Column(db.Boolean, default=True, nullable=False)

    # Relationship to User
    user = db.relationship('User', backref=db.backref('sessions', lazy=True))

    def __repr__(self):
        return '<UserSession %r>' % self.token



import json
import os
import struct
import numpy as np

'''
    Load .utk and return a json that represents a layer
'''
def load_utk(filepath):

    file = open(filepath, mode='r')
    file_content = json.loads(file.read())

    directory = os.path.dirname(filepath)
    # file name with extension
    file_name = os.path.basename(filepath)
    # file name without extension
    file_name_wo_extension = os.path.splitext(file_name)[0]

    coordinates = []
    normals = []
    indices = []
    ids = []

    if('coordinates' in file_content['data'][0]['geometry']):
        f = open(os.path.join(directory,file_name_wo_extension+'_coordinates.data'), "rb")

        data = f.read()

        unpacked_data = struct.iter_unpack('d', data)

        for elem in unpacked_data:
            coordinates.append(elem[0])

        f.close()
    if('normals' in file_content['data'][0]['geometry']):
        f = open(os.path.join(directory,file_name_wo_extension+'_normals.data'), "rb")

        data = f.read()

        unpacked_data = struct.iter_unpack('f', data)

        for elem in unpacked_data:
            normals.append(elem[0])

        f.close()
    if('indices' in file_content['data'][0]['geometry']):
        f = open(os.path.join(directory,file_name_wo_extension+'_indices.data'), "rb")

        data = f.read()

        unpacked_data = struct.iter_unpack('I', data)

        for elem in unpacked_data:
            indices.append(elem[0])

        f.close()
    if('ids' in file_content['data'][0]['geometry']):
        f = open(os.path.join(directory,file_name_wo_extension+'_ids.data'), "rb")

        data = f.read()

        unpacked_data = struct.iter_unpack('I', data)

        for elem in unpacked_data:
            ids.append(elem[0])

        f.close()

    for i in range(len(file_content['data'])):

        if(len(coordinates) > 0):
            startAndSize = file_content['data'][i]['geometry']['coordinates']
            file_content['data'][i]['geometry']['coordinates'] = coordinates[startAndSize[0]:startAndSize[0]+startAndSize[1]]

        if(len(indices) > 0):
            startAndSize = file_content['data'][i]['geometry']['indices']
            file_content['data'][i]['geometry']['indices'] = indices[startAndSize[0]:startAndSize[0]+startAndSize[1]]

        if(len(normals) > 0):
            startAndSize = file_content['data'][i]['geometry']['normals']
            file_content['data'][i]['geometry']['normals'] = normals[startAndSize[0]:startAndSize[0]+startAndSize[1]]

        if(len(ids) > 0):
            startAndSize = file_content['data'][i]['geometry']['ids']
            file_content['data'][i]['geometry']['ids'] = ids[startAndSize[0]:startAndSize[0]+startAndSize[1]]

    file.close()

    return file_content

'''
    Get all coordinates (in a flat array) of a json layer
'''
def get_coordinates(layer_json):
    
    if(len(layer_json['data']) == 0):
        return []
    else:
        if 'coordinates' not in layer_json['data'][0]['geometry']:
            raise Exception('Layer does not have a coordinates field')

    coordinates = []

    for geometry in layer_json['data']:
        coordinates += geometry['geometry']['coordinates']

    return coordinates

'''
    Get all indices (in a flat array) of a json layer
'''
def get_indices(layer_json):
    
    if(len(layer_json['data']) == 0):
        return []
    else:
        if 'indices' not in layer_json['data'][0]['geometry']:
            raise Exception('Layer does not have a indices field')

    indices = []

    for geometry in layer_json['data']:
        indices += geometry['geometry']['indices']

    return indices

'''
    Get all normals (in a flat array) of a json layer
'''
def get_normals(layer_json):
    
    if(len(layer_json['data']) == 0):
        return []
    else:
        if 'normals' not in layer_json['data'][0]['geometry']:
            raise Exception('Layer does not have a normals field')

    normals = []

    for geometry in layer_json['data']:
        normals += geometry['geometry']['normals']

    return normals

'''
    Get all ids (in a flat array) of a json layer
'''
def get_ids(layer_json):
    
    if(len(layer_json['data']) == 0):
        return []
    else:
        if 'ids' not in layer_json['data'][0]['geometry']:
            raise Exception('Layer does not have a ids field')

    ids = []

    for geometry in layer_json['data']:
        ids += geometry['geometry']['ids']

    return ids


import unittest
import tempfile
import os
import sys
import json
from flask import Flask, jsonify
from utk_curio.backend.app.api.routes import bp

# Initialize the Flask app for testing
app = Flask(__name__)
app.register_blueprint(bp)

# Modify sys.path to include the backend folder
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'backend')))

test_data = {
    "workflow": "test_workflow",
    "data": {
        "workflow_name": "test_workflow",  # Make sure this matches the workflow name inserted above
        "activity_name": "COMPUTATION_ANALYSIS",  # Include the activity name in the request
        "activityexec_start_time": "2025-05-09T02:00:00",
        "activityexec_end_time": "2025-05-09T03:00:00",
        "types_input": {"input_relation_1": 1},  # Sample input data
        "types_output": {"output_relation_1": 1},  # Sample output data
        "activity_source_code": "return sum(args[0]) / len(args[0])",
        "input": {
            "dataType": "dataframe",
            "path": "./tests/test.data"  
        }
    }
}

class TestRoutes(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.client = app.test_client()

    def test_live(self):
        response = self.client.get('/live')
        self.assertEqual(response.data.decode('utf-8'), 'Backend is live.')
        self.assertEqual(response.status_code, 200)

    def test_upload_file(self):
        # Create a temporary file for the test
        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(b"Test content")  # Write some test data to the file
            tmp_file.seek(0)  # Move the file pointer back to the beginning for reading
            temp_file_path = tmp_file.name  # Save the file path to use in the test

        try:
            # Simulate a file upload using the temporary file
            with open(temp_file_path, 'rb') as file:
                data = {
                    'file': (file, 'test_file.txt')  # Simulate the file upload
                }
                response = self.client.post('/upload', data=data)
                
                # Test if the file upload was successful
                self.assertIn('File uploaded successfully', response.data.decode('utf-8'))
                self.assertEqual(response.status_code, 200)

        finally:
            # Clean up: Delete the temporary file after the test
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_process_python_code(self):
        test_code = {
            "code": test_data["data"]["activity_source_code"],
            "boxType": test_data["data"]["activity_name"],
            "input": {
                "dataType": test_data["data"]["input"]["dataType"],
                "path": test_data["data"]["input"].get("path", ""),
                "data": test_data["data"]["input"].get("data", "")
            }
        }
        response = self.client.post('/processPythonCode', json=test_code)
        self.assertEqual(response.status_code, 200)

    def test_db(self):
        response = self.client.get('/checkDB')
        self.assertEqual(response.status_code, 200)

    def test_prov(self):

        response = self.client.get('/truncateDBProv')
        self.assertEqual(response.status_code, 200)

        response = self.client.post('/saveWorkflowProv', json=test_data)
        self.assertEqual(response.status_code, 200)
    
        response = self.client.post('/newBoxProv', json=test_data)
        self.assertEqual(response.status_code, 200)

        # response = self.client.post('/boxExecProv', json=test_data)
        # self.assertEqual(response.status_code, 200)

        response = self.client.post('/getBoxGraph', json=test_data)
        self.assertEqual(response.status_code, 200)

        response = self.client.post('/deleteBoxProv', json=test_data)
        self.assertEqual(response.status_code, 200)


if __name__ == "__main__":
    unittest.main()